//---------------------------------------------------------------------------
/*
	Risa [肳]      alias gg3 [kirikiri-3]
	 stands for "Risa Is a Stagecraft Architecture"
	Copyright (C) 2000 W.Dee <dee@kikyou.info> and contributors

	See details of license at "license.txt"
*/
//---------------------------------------------------------------------------
//! @file
//! @brief Oobt@鎩Ɛev[gNX
//---------------------------------------------------------------------------
#ifndef RingBufferH
#define RingBufferH

#include <stddef.h>
/*
	Oobt@, ring buffer, circular buffer, obt@
*/

//---------------------------------------------------------------------------
//! @brief		Œ蒷Oobt@̎
//---------------------------------------------------------------------------
template <typename T>
class tRisaRingBuffer
{
	T * Buffer; //!< obt@
	size_t Size; //!< obt@̃TCY
	size_t WritePos; //!< ݈ʒu
	size_t ReadPos; //!< ǂݍ݈ʒu
	size_t DataSize; //!< obt@ɓĂf[^̃TCY

public:
	//! @brief RXgN^
	tRisaRingBuffer(size_t size)
	{
		Size = size;
		Buffer = new T[Size];
		WritePos = ReadPos = 0;
		DataSize = 0;
	}

	//! @brief fXgN^
	~tRisaRingBuffer()
	{
		delete [] Buffer;
	}

	//! @brief	TCY𓾂
	size_t GetSize() { return Size; }

	//! @brief	݈ʒu𓾂
	size_t GetWritePos() { return WritePos; }

	//! @brief	ǂݍ݈ʒu𓾂
	size_t GetReadPos() { return ReadPos; }

	//! @brief	obt@ɓĂf[^̃TCY𓾂
	size_t GetDataSize() { return DataSize; }

	//! @brief	obt@̋󂫗eʂ𓾂
	size_t GetFreeSize() { return Size - DataSize; }

	//! @brief	obt@ǂݍނ߂̃|C^𓾂
	//! @param	readsize ǂݍ݂f[^ ( 1 ȏ̐; 0 nȂ )
	//! @param	p1		ubN1̐擪ւ̃|C^i[邽߂̕ϐ
	//! @param	p1size	p1̕\ubÑTCY
	//! @param	p2		ubN2̐擪ւ̃|C^i[邽߂̕ϐ(NULL蓾)
	//! @param	p2size	p2̕\ubÑTCY(0蓾)
	//! @param	offset	ReadPos ɉZItZbg
	//! @note	obt@ƂĂAۂ̓jAȗ̈Ƀobt@mۂĂB
	//!			̂߁A ReadPos + readsize obt@̏I[𒴂ĂꍇA
	//!			ubN͂Qɕf邱ƂɂȂB
	//!			̃\bh́AreadsizeۂɃobt@ɓĂf[^̃TCYȉł邩
	//!			Ȃǂ̃`FbN͂sȂBO GetDataSize 𒲂ׁAǂݍ݂
	//!			TCYۂɃobt@ɂ邩ǂ`FbN邱ƁB
	void GetReadPointer(size_t readsize,
						const T * & p1, size_t &p1size,
						const T * & p2, size_t &p2size,
						ptrdiff_t offset = 0)
	{
		size_t pos = ReadPos + offset;
		while(pos >= Size) pos -= Size;
		if(readsize + pos > Size)
		{
			// readsize + pos obt@̏I[𒴂Ă
			//   ԂubN2
			p1 = pos + Buffer;
			p1size = Size - pos;
			p2 = Buffer;
			p2size = readsize - p1size;
		}
		else
		{
			// readsize + pos obt@̏I[𒴂ĂȂ
			//   ԂubN1
			p1 = pos + Buffer;
			p1size = readsize;
			p2 = NULL;
			p2size = 0;
		}
	}

	//! @brief	ǂݍ݃|C^i߂
	//! @param	advance		i߂vf
	//! @note	̃\bh͎ۂ advance < GetDataSize() ł邱ƂmFȂB
	//!			KvȂΌĂяoŃ`FbN邱ƁB
	void AdvanceReadPos(size_t advance = 1)
	{
		ReadPos += advance;
		if(ReadPos >= Size) ReadPos -= Size;
		DataSize -= advance;
	}

	//! @brief	ŏ̗vfԂ
	//! @return	ŏ̗vfւ̎Q
	//! @note	ŏ̗vfւ̎QƂAĂBvfobt@ɖƂ͖ȗvf
	//!			(ANZXłȂvf)AĂ̂ŁAOɃobt@ɗvf1ȏ
	//!			݂邱ƂmF邱ƁB̃\bh͓ǂݍ݃|C^ړȂB
	const T & GetFirst() const
	{
		size_t pos = ReadPos;
		return Buffer[pos];
	}

	//! @brief	nԖڂ̗vfԂ
	//! @return	nԖڂ̗vfւ̎Q
	//! @note	nԖڂ̗vfւ̎QƂAĂBvfobt@ɖƂ͈͊O̎
	//!			͖̓`łB̃\bh͓ǂݍ݃|C^ړȂB
	const T & GetAt(size_t n) const
	{
		size_t pos = ReadPos + n;
		while(pos >= Size) pos -= Size;
		return Buffer[pos];
	}

	//! @brief	obt@ɏނ߂̃|C^𓾂
	//! @param	writesize ݂f[^ ( 1 ȏ̐; 0 nȂ )
	//! @param	p1		ubN1̐擪ւ̃|C^i[邽߂̕ϐ
	//! @param	p1size	p1̕\ubÑTCY
	//! @param	p2		ubN2̐擪ւ̃|C^i[邽߂̕ϐ(NULL蓾)
	//! @param	p2size	p2̕\ubÑTCY(0蓾)
	//! @param	offset	WritePos ɉZItZbg
	//! @note	GetReadPointer̐QƂ̂
	void GetWritePointer(size_t writesize,
						T * & p1, size_t &p1size,
						T * & p2, size_t &p2size,
						ptrdiff_t offset = 0)
	{
		size_t pos = WritePos + offset;
		while(pos >= Size) pos -= Size;
		if(writesize + pos > Size)
		{
			// writesize + pos obt@̏I[𒴂Ă
			//   ԂubN2
			p1 = pos + Buffer;
			p1size = Size - pos;
			p2 = Buffer;
			p2size = writesize - p1size;
		}
		else
		{
			// writesize + pos obt@̏I[𒴂ĂȂ
			//   ԂubN1
			p1 = pos + Buffer;
			p1size = writesize;
			p2 = NULL;
			p2size = 0;
		}
	}

	//! @brief	݃|C^i߂
	//! @param	advance		i߂vf
	//! @note	̃\bh͎ۂ advance < GetFreeSize() ł邱ƂmFȂB
	//!			KvȂΌĂяoŃ`FbN邱ƁB
	void AdvanceWritePos(size_t advance = 1)
	{
		WritePos += advance;
		if(WritePos >= Size) WritePos -= Size;
		DataSize += advance;
	}

	//! @brief	݃|C^i߁Aobt@ӂꂽ擪̂Ă
	//! @param	advance		i߂vf
	//! @note	AdvanceWritePos ƈقȂAobt@ӂꂽAf[^̐擪̂ĂB
	void AdvanceWritePosWithDiscard(size_t advance = 1)
	{
		WritePos += advance;
		if(WritePos >= Size) WritePos -= Size;
		DataSize += advance;
		if(DataSize > Size)
		{
			AdvanceReadPos(DataSize - Size);
		}
	}

	//! @brief	݈ʒu̗vfԂ
	//! @return	݈ʒu̗vfւ̎Q
	//! @note	݈ʒu̗vfւ̎QƂAĂB̃\bh̓obt@ɋ
	//!			邩ǂ̃`FbN͍sȂ̂Œӂ邱ƁB
	//!			̃\bh̓obt@݈̏ʒuړȂB
	T & GetLast()
	{
		return Buffer[WritePos];
	}
};

#endif
